import os
from pathlib import Path
from shutil import copyfile
from typing import Dict, Union

import yaml
from fontawesome import icons
from i3ipc.aio import Con
from xdg import XDG_CONFIG_HOME

POSSIBLE_SWAY_CONFIG_PATHS = ['sway', 'i3']

CONFIG_FILE_NAME = 'sdn-config.yaml'

DEFAULT_DELIMITER = "|"
DEFAULT_DEFAULT_ICON = "dot-circle"


class ConfigException(Exception):
    pass


class ClientConfig:
    def __init__(self, key: str, data: Union[str, Dict]):
        self.key = key
        if type(data) == str:
            self.icon = icons.get(data, data)
        elif type(data) == dict:
            self.icon = icons.get(data['icon'], data['icon'])
        else:
            raise ConfigException(f'clients/{key}: invalid entity {data}')

    def get_symbol(self, leaf: Con):
        return self.icon


class Config:
    _client_configs: Dict[str, ClientConfig] = {}
    _delimiter: str = "|"
    _last_modified: float = None

    def __init__(self, use_default=False):
        self.config_location = Config._find_config() if not use_default else Config._default_config_path()
        print(f"Using config file at {self.config_location}")

    def _load(self):
        if os.path.getmtime(self.config_location) != self._last_modified:
            with open(self.config_location, 'r') as f:
                data = yaml.safe_load(f)
            self._last_modified = os.path.getmtime(self.config_location)

            self._client_configs = {k: ClientConfig(k, v) for k, v in data.get('clients', {}).items()}
            self._delimiter = data.get('deliminator', DEFAULT_DELIMITER)
            self._default_icon = data.get('default_icon', DEFAULT_DEFAULT_ICON)
            self._default_icon = icons.get(self._default_icon, self._default_icon)

    @property
    def client_configs(self):
        self._load()
        return self._client_configs

    @property
    def delimiter(self):
        self._load()
        return self._delimiter

    @property
    def default_icon(self):
        self._load()
        return self._default_icon

    @staticmethod
    def _find_config():
        for possible_path in [pp.joinpath(CONFIG_FILE_NAME) for pp in Config._find_sway_folders()]:
            if possible_path.exists():
                return possible_path
        print(Config._find_sway_folders())
        return Config._create_config(Config._find_sway_folders()[0].joinpath(CONFIG_FILE_NAME))

    @staticmethod
    def _find_sway_folders():
        possible_paths = [XDG_CONFIG_HOME.joinpath(pp) for pp in POSSIBLE_SWAY_CONFIG_PATHS]
        return [pp for pp in possible_paths if pp.exists()]

    @staticmethod
    def _create_config(config_location: str):
        print(f"Creating default config file at {config_location}")
        copyfile(Config._default_config_path(), config_location)
        return config_location

    @staticmethod
    def _default_config_path():
        return Path(os.path.realpath(__file__)).parent.joinpath('default.yaml')



from setuptools import setup, find_packages

with open('README.md') as f:
    long_description = f.read()

setup(name='sway-dynamic-names',
      description='Dynamically update the name of each Sway WM workspace using font-awesome icons',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://github.com/j-waters/sway-dynamic-workspace-names',
      license='MIT',
      install_requires=["i3ipc", "pyyaml", "fontawesome", "xdg", "Click"],
      author='James Waters',
      author_email='james@jcwaters.co.uk',
      packages=find_packages(),
      include_package_data=True,
      entry_points={
          'console_scripts': [
              'sway-dynamic-names=sway_dynamic_names.__main__:main'
          ]
      },
      setup_requires=['setuptools-git-versioning'],
      version_config={
          "template": "{tag}",
          "dev_template": "{tag}.dev{ccount}+git.{sha}",
          "dirty_template": "{tag}.dev{ccount}+git.{sha}.dirty",
          "starting_version": "0.0.1",
          "version_file": "",
          "count_commits_from_version_file": False
      },
      python_requires='>=3.8',
      )



import re

from i3ipc import Event
from i3ipc.aio import Connection, Con
from i3ipc.events import IpcBaseEvent

from .config import Config


class Watcher:
    i3: Connection

    def __init__(self, config: Config):
        self.config = config
        self.commands = []

    async def start(self):
        self.i3 = await Connection().connect()
        for case in [Event.WINDOW_MOVE, Event.WINDOW_NEW, Event.WINDOW_TITLE, Event.WINDOW_CLOSE]:
            self.i3.on(case, self.callback)
        await self.rename_all()
        await self.i3.main()

    async def callback(self, conn: Connection, event: IpcBaseEvent):
        await self.rename_all()

    async def rename_all(self):
        tree = await self.i3.get_tree()
        workspaces = tree.workspaces()
        for num, workspace in enumerate(workspaces):
            await self.rename_workspace(workspace, num)
        await self.commit()

    async def rename_workspace(self, workspace: Con, num: int):
        symbols = []
        for leaf in workspace.leaves():
            leaf_symbol = self.get_symbol(leaf)
            if leaf_symbol:
                symbols.append(leaf_symbol)
        new_name = self.config.default_icon
        if symbols:
            new_name = self.config.delimiter.join(symbols)
        new_name = f"{num}:{new_name}"
        if workspace.name != new_name:
            workspace_name_san = workspace.name.replace('"', '\\"')
            new_name_san = new_name.replace('"', '\\"')
            self.commands.append(f'rename workspace "{workspace_name_san}" to "{new_name_san}"')

    def get_symbol(self, leaf: Con):
        for identifier in ('name', 'window_title', 'window_instance', 'window_class'):
            name = getattr(leaf, identifier, None)
            if name is None:
                continue
            for name_re, conf in self.config.client_configs.items():
                if re.match(name_re, name, re.IGNORECASE):
                    return conf.get_symbol(leaf)

    async def commit(self):
        await self.i3.command(u';'.join(self.commands))
        self.commands = []


async def start():
    config = Config(use_default=True)
    watcher = Watcher(config)
    await watcher.start()
